<a class="sidebar-brand{% if logo %} centered{% endif %}" href="{{ pathto(master_doc) }}">
    {% block brand_content %}
    {%- if logo %}
    <div class="sidebar-logo-container">
        <img class="sidebar-logo" src="{{ pathto('_static/' + logo, 1) }}" alt="Logo" />
    </div>
    {%- endif %}
    {%- if theme_light_logo and theme_dark_logo %}
    <div class="sidebar-logo-container">
        <img class="sidebar-logo only-light" src="{{ pathto('_static/' + theme_light_logo, 1) }}" alt="Light Logo" />
        <img class="sidebar-logo only-dark" src="{{ pathto('_static/' + theme_dark_logo, 1) }}" alt="Dark Logo" />
    </div>
    {%- endif %}
    {% if not theme_sidebar_hide_name %}
    <span class="sidebar-brand-text">{{ project }}
        <small>{{current_version_tag}}</small></span>
    {%- endif %}
    {% endblock brand_content %}
</a>