{%- if current_version %}
<div class="rst-versions">

    <div class="rst-other-versions">
        {%- if versions.branches %}
        <dl>
            <dt>Versions</dt>
            {%- for item in versions %}
            {%- if not current_version.name == item.name %}
            <dd><a href="{{ item.url }}">{{ item.version }}</a></dd>
            {%- else %}
            <dd> {{ item.version }} </dd>
            {%- endif %}
            {%- endfor %}
        </dl>
        {%- endif %}
    </div>
</div>
{%- endif %}